# -*- coding: utf-8 -*-

# Le but de ce programme est de comparer plusieurs méthodes d'approximations numériques d'intégrales sur un segment.

# 1. La première méthode utilise des sommes de Riemann classiques (le point auquel la fonction f est évaluée est à gauche de l'intervalle) :
# ⁽ᵇ⁻ᵃ⁾⁄ₙ ∑ₖ₌₀ⁿ⁻¹ f(a+k⁽ᵇ⁻ᵃ⁾⁄ₙ) converge vers ∫f(t)dt.
# Pour une fonction de classe C¹, la vitesse de convergence est en O(¹⁄ₙ).

# 2. La seconde méthode utilise des sommes de Riemann au point milieu (le point auquel la fonction f est évaluée est au milieu de l'intervalle) :
# ⁽ᵇ⁻ᵃ⁾⁄ₙ ∑ₖ₌₀ⁿ⁻¹ f(a+(k+½)⁽ᵇ⁻ᵃ⁾⁄ₙ) converge vers ∫f(t)dt.
# Pour une fonction de classe C², la vitesse de convergence est en O(¹⁄ₙ²).

# 3. La troisième méthode, ajoutée pour compléter ce panorama, est la méthode de Simpson. 
# On évalue f à la fois au milieu, à gauche et à droite de chaque intervalle, 
# et on pondère les valeurs obtenues pour éliminer tous les termes d'ordre au plus 3.
# ⁽ᵇ⁻ᵃ⁾⁄ₙ ∑ₖ₌₀ⁿ⁻¹ ⅙[f(a+k⁽ᵇ⁻ᵃ⁾⁄ₙ)+4f(a+(k+½)⁽ᵇ⁻ᵃ⁾⁄ₙ)+f(a+(k+1)⁽ᵇ⁻ᵃ⁾⁄ₙ)] converge vers ∫f(t)dt.
# Pour une fonction de classe C⁴, la vitesse de convergence est en O(¹⁄ₙ⁴)

# 4. Pour comparer ces méthodes, on se donne une intégrale connue :
# ∫₋₁¹ eᵗ dt = e-¹⁄ₑ.
# On va donc utiliser ces méthodes d'intégration avec f la fonction exponentielle, a=-1, b=1.
# En prenant des valeurs de n élevées, on pourra estimer l'efficacité de ces différentes méthodes.
# N'hésitez pas à tester ce programme en modifiant la fonction et les bornes d'intégration !


import numpy

# La première fonction prend en entrée deux réels (float) a et b et un entier n
# Elle ressort les sommes de Riemann selon la première méthode de la fonction exponentielle entre a et b.

def somme_Riemann_point_gauche (a, b, n) :
    pas = (b-a)/n # On pré-calcule le pas. Cela clarifie les calculs suivants.
    S = 0
    for k in range(n) : # Attention : range(n) contient exactement les entiers de 0 à n-1. C'est ce que l'on veut.
        S += numpy.exp(a+k*pas)
    return pas*S


# La deuxième fonction prend en entrée deux réels (float) a et b et un entier n
# Elle ressort les sommes de Riemann selon la deuxième méthode de la fonction exponentielle entre a et b.

def somme_Riemann_point_milieu (a, b, n) :
    pas = (b-a)/n
    S = 0
    for k in range(n) :
        S += numpy.exp(a+(k+0.5)*pas)
    return pas*S


# La troisième fonction prend en entrée deux réels (float) a et b et un entier n
# Elle ressort les sommes de Riemann selon la troisième méthode de la fonction exponentielle entre a et b.

def somme_Riemann_methode_Simpson (a, b, n) :
    pas = (b-a)/n
    S = 0
    for k in range(n) :
        S += numpy.exp(a+k*pas) + 4*numpy.exp(a+(k+0.5)*pas) + numpy.exp(a+(k+1)*pas)
    return (pas/6)*S


# Valeur attendue : e-¹⁄ₑ
cible = numpy.exp(1)-numpy.exp(-1)


# On imprime les valeurs données par ces trois méthodes dans l'ordre. La valeur de n est à ajuster à volonté.
n = 10**6
print('Erreur commise avec la méthode classique :', abs(somme_Riemann_point_gauche(-1,1,n)-cible))
print('Erreur commise avec la méthode du point milieu :', abs(somme_Riemann_point_milieu(-1,1,n)-cible))
print('Erreur commise avec la méthode de Simpson :', abs(somme_Riemann_methode_Simpson(-1,1,n)-cible))


# Les ordres de grandeurs des erreurs des trois méthodes sont ceux annoncés.
# Pour de grandes valeurs de n, la méthode de Simpson a tendance à stagner.
# Dans ce régime, les erreurs sont majoritairement dues à la précision des réels (float) 
# et à l'évaluation de la fonction exp, erreurs qui se cumulent quand on additionne de nombreux termes.